# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import logging
import sys

from os.path import exists

from bumpr.helpers import execute, BumprError

if sys.version_info < (3, 0):
    from io import open

logger = logging.getLogger(__name__)

__all_ = (
    'Hook',
    'ReadTheDocHook',
    'ChangelogHook',
    'CommandHook',
    'ReplaceHook',
    'HOOKS',
)


class Hook(object):
    key = None
    defaults = None

    def __init__(self, releaser):
        self.releaser = releaser
        self.verbose = releaser.config.verbose
        self.dryrun = releaser.config.dryrun
        self.config = releaser.config[self.key]
        self.validate()

    def validate(self):
        '''Override this method to implement initial validation'''

    def bump(self, replacements):
        pass

    def prepare(self, replacements):
        pass


class ReadTheDocHook(Hook):
    '''
    This hook set the readthedoc url corresponding to the version
    '''
    key = 'readthedoc'
    defaults = {
        'id': None,
        'url': 'https://{id}.readthedocs.io/en/{tag}',
        'badge': 'https://readthedocs.org/projects/{id}/badge/?version={tag}',
        'bump': '{version}',
        'prepare': 'latest',
    }

    def url(self, tag):
        return self.config.url.format(id=self.config.id, tag=tag)

    def badge(self, tag):
        return self.config.badge.format(id=self.config.id, tag=tag)

    def bump(self, replacements):
        replacements.insert(0, (
            self.badge('latest'),
            self.badge(self.releaser.version)
        ))
        replacements.insert(0, (
            self.url('latest'),
            self.url(self.releaser.version)
        ))

    def prepare(self, replacements):
        replacements.insert(0, (
            self.badge(self.releaser.version),
            self.badge('latest')
        ))
        replacements.insert(0, (
            self.url(self.releaser.version),
            self.url('latest')
        ))


class ChangelogHook(Hook):
    '''
    This hook bump the changelog version header and prepare a new section for the next release.
    '''
    key = 'changelog'
    defaults = {
        'file': None,
        'separator': '-',
        'bump': '{version} ({date:%Y-%m-%d})',
        'prepare': 'Current',
        'empty': 'Nothing yet',
    }

    def validate(self):
        if not self.config.get('file'):
            raise BumprError('Changelog file has not been specified')
        elif not exists(self.config.file):
            raise BumprError('Changelog file does not exists')

    def bump(self, replacements):
        with open(self.config.file, 'r', encoding=self.releaser.config.encoding) as changelog_file:
            before = changelog_file.read()
            after = before.replace(self.dev_header(), self.bumped_header())
        self.releaser.perform(self.config.file, before, after)

    def prepare(self, replacements):
        next_header = '\n'.join((self.dev_header(), '', '- {0}'.format(self.config.empty), '', self.bumped_header()))
        with open(self.config.file, 'r', encoding=self.releaser.config.encoding) as changelog_file:
            before = changelog_file.read()
            after = before.replace(self.bumped_header(), next_header)
        self.releaser.perform(self.config.file, before, after)

    def dev_header(self):
        return self.underline(self.config.prepare)

    def bumped_header(self):
        title = self.config.bump.format(
            version=self.releaser.version,
            date=self.releaser.timestamp,
            **self.releaser.version.__dict__
        )
        return self.underline(title)

    def underline(self, text):
        if self.config.separator:
            return '\n'.join((text, len(text) * self.config.separator))
        else:
            return text


class CommandsHook(Hook):
    '''
    This hook execute commands
    '''
    key = 'commands'
    defaults = {
        'bump': None,
        'prepare': None,
    }

    def bump(self, replacements):
        if self.config.bump:
            replacements = dict(
                version=self.releaser.version,
                date=self.releaser.timestamp,
                **self.releaser.version.__dict__
            )
            execute(self.config.bump, replacements=replacements, verbose=self.verbose, dryrun=self.dryrun)

    def prepare(self, replacements):
        if self.config.prepare:
            replacements = dict(
                version=self.releaser.next_version,
                date=self.releaser.timestamp,
                **self.releaser.next_version.__dict__
            )
            execute(self.config.prepare, replacements=replacements, verbose=self.verbose, dryrun=self.dryrun)


class ReplaceHook(Hook):
    '''
    This hook perform replacements in files
    '''
    key = 'replace'
    defaults = {}

    def bump(self, replacements):
        replacements.insert(0, (
            self.config.dev.format(version=self.releaser.prev_version,
                                   date=self.releaser.timestamp,
                                   **self.releaser.prev_version.__dict__),
            self.config.stable.format(version=self.releaser.version,
                                      date=self.releaser.timestamp,
                                      **self.releaser.version.__dict__)
        ))

    def prepare(self, replacements):
        replacements.insert(0, (
            self.config.stable.format(version=self.releaser.version,
                                      date=self.releaser.timestamp,
                                      **self.releaser.version.__dict__),
            self.config.dev.format(version=self.releaser.next_version,
                                   date=self.releaser.timestamp,
                                   **self.releaser.next_version.__dict__)
        ))


HOOKS = (ReadTheDocHook, ChangelogHook, CommandsHook, ReplaceHook)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Bump'R: Version bumper and Python package releaser

- Clean-up release artifact
- Bump version and tag it
- Build a source distrbution and upload on PyPI
- Update version for new develpoment cycle
- Can run test suite before
- Can be customized with a config file
- Extensible with hooks
'''

from .__about__ import __description__, __version__  # noqa


def main():
    import sys
    from bumpr import log
    log.init()

    from bumpr.config import Config, ValidationError
    from bumpr.releaser import Releaser
    from bumpr.helpers import BumprError
    from logging import DEBUG, INFO, getLogger

    config = Config.parse_args()
    getLogger().setLevel(DEBUG if config.verbose else INFO)
    logger = getLogger(__name__)

    try:
        config.validate()
    except ValidationError as e:
        msg = 'Invalid configuration: {0}'.format(e)
        logger.error(msg)
        sys.exit(1)

    try:
        releaser = Releaser(config)
        releaser.release()
    except BumprError as error:
        logger.error(str(error))
        sys.exit(1)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import logging
import re
import sys

from datetime import datetime
from difflib import unified_diff

from bumpr.helpers import execute, BumprError
from bumpr.hooks import HOOKS
from bumpr.vcs import VCS
from bumpr.version import Version

if sys.version_info < (3, 0):
    from io import open

logger = logging.getLogger(__name__)


class Releaser(object):
    '''
    Release workflow executor
    '''
    def __init__(self, config):
        self.config = config

        with open(config.file) as f:
            match = re.search(config.regex, f.read())
            try:
                version_string = match.group('version')
                self.prev_version = Version.parse(version_string)
            except:
                raise BumprError('Unable to extract version from {0}'.format(config.file))

        logger.debug('Previous version: {0}'.format(self.prev_version))

        self.version = self.prev_version.copy()
        self.version.bump(config.bump.part, config.bump.unsuffix, config.bump.suffix)
        logger.debug('Bumped version: {0}'.format(self.version))

        self.next_version = self.version.copy()
        self.next_version.bump(config.prepare.part, config.prepare.unsuffix, config.prepare.suffix)
        logger.debug('Prepared version: {0}'.format(self.next_version))

        self.timestamp = None

        if config.vcs:
            self.vcs = VCS[config.vcs](verbose=config.verbose)
            self.vcs.validate()

        if config.dryrun:
            self.diffs = {}

        self.hooks = [hook(self) for hook in HOOKS if self.config[hook.key]]

    def execute(self, command, version=None, verbose=None):
        version = version or self.version
        verbose = verbose or self.config.verbose
        replacements = dict(version=version, date=self.timestamp, **version.__dict__)
        execute(command, replacements=replacements, dryrun=self.config.dryrun, verbose=verbose)

    def release(self):
        self.timestamp = datetime.now()

        if self.config.bump_only:
            self.bump()
        elif self.config.prepare_only:
            self.prepare()
        else:
            self.clean()
            self.test()
            self.bump()
            self.publish()
            self.prepare()
            self.push()

    def test(self):
        if self.config.tests:
            logger.info('Running test suite')
            self.execute(self.config.tests, verbose=True)

    def bump(self):
        logger.info('Bump version %s', self.version)

        replacements = [
            (str(self.prev_version), str(self.version))
        ]

        for hook in self.hooks:
            hook.bump(replacements)

        self.bump_files(replacements)

        if self.config.vcs:
            self.commit(self.config.bump.message.format(
                version=self.version,
                date=self.timestamp,
                **self.version.__dict__
            ))
            self.tag()

        if self.config.dryrun:
            self.display_diff()
            self.diffs.clear()

    def prepare(self):
        if self.version == self.next_version:
            logger.info('Skip prepare phase')
            return
        logger.info('Prepare version %s', self.next_version)

        replacements = [
            (str(self.version), str(self.next_version))
        ]

        for hook in self.hooks:
            hook.prepare(replacements)

        self.bump_files(replacements)

        if self.config.vcs:
            self.commit(self.config.prepare.message.format(
                version=self.next_version,
                date=self.timestamp,
                **self.next_version.__dict__
            ))

        if self.config.dryrun:
            self.display_diff()

    def clean(self):
        '''Clean the workspace'''
        if self.config.clean:
            logger.info('Cleaning')
            self.execute(self.config.clean)

    def perform(self, filename, before, after):
        if before == after:
            return
        if self.config.dryrun:
            diff = unified_diff(before.split('\n'), after.split('\n'), lineterm='')
            self.diffs[filename] = diff
        else:
            with open(filename, 'w', encoding=self.config.encoding) as f:
                f.write(after)

    def bump_files(self, replacements):
        for filename in [self.config.file] + self.config.files:
            with open(filename, 'r', encoding=self.config.encoding) as current_file:
                before = current_file.read()
            after = before
            for token, replacement in replacements:
                after = after.replace(token, replacement)
            self.perform(filename, before, after)

    def publish(self):
        '''Publish the current release to PyPI'''
        if self.config.publish:
            logger.info('Publish')
            self.execute(self.config.publish)

    def tag(self):
        if self.config.commit and self.config.tag:
            logger.debug('Tag: %s', self.version)
            if not self.config.dryrun:
                self.vcs.tag(str(self.version))
            else:
                logger.dryrun('tag: {0}'.format(self.version))

    def commit(self, message):
        if self.config.commit:
            logger.debug('Commit: %s', message)
            if not self.config.dryrun:
                self.vcs.commit(message)
            else:
                logger.dryrun('commit: {0}'.format(message))

    def push(self):
        if self.config.vcs and self.config.commit and self.config.push:
            logger.info('Push to upstream repository')
            if not self.config.dryrun:
                self.vcs.push()
            else:
                logger.dryrun('push to remote repository')

    def display_diff(self):
        for filename, diff in self.diffs.items():
            logger.diff(filename)
            for line in diff:
                logger.diff(line)
            logger.diff('')



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import logging
import shlex
import subprocess
import sys

IS_PY26 = sys.version_info[:2] == (2, 6)


class BumprError(Exception):
    pass


def check_output(*args, **kwargs):
    '''Compatibility wrapper for Python 2.6 missin g subprocess.check_output'''
    if hasattr(subprocess, 'check_output'):
        return subprocess.check_output(stderr=subprocess.STDOUT, universal_newlines=True,
                                       *args, **kwargs)
    else:
        process = subprocess.Popen(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                   universal_newlines=True, **kwargs)
        output, _ = process.communicate()
        retcode = process.poll()
        if retcode:
            error = subprocess.CalledProcessError(retcode, args[0])
            error.output = output
            raise error
        return output


def execute(command, verbose=False, replacements=None, dryrun=False):
    logger = logging.getLogger(__name__)
    replacements = replacements or {}
    if not command:
        return
    elif isinstance(command, (list, tuple)):
        if not isinstance(command[0], (list, tuple)):
            command = [command]
        commands = []
        for cmd in command:
            commands.append([part.format(**replacements) for part in cmd])
    else:
        command = command.format(**replacements)
        if IS_PY26:
            command = command.encode('utf8')
        commands = [shlex.split(cmd.strip()) for cmd in command.splitlines() if cmd.strip()]

    output = ''
    for cmd in commands:
        try:
            if dryrun:
                logger.dryrun('execute: {0}'.format(' '.join(cmd)))
            elif verbose:
                subprocess.check_call(cmd)
            else:
                output += check_output(cmd)
        except subprocess.CalledProcessError as exception:
            if hasattr(exception, 'output') and exception.output:
                print(exception.output)
            cmd = ' '.join(cmd) if isinstance(cmd, (list, tuple)) else cmd
            raise BumprError('Command "{0}" failed with exit code {1}'.format(cmd, exception.returncode))
    return output


class ObjectDict(dict):
    '''A dictionnary with object-like attribute access and depp merge'''
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, ObjectDict):
            value = ObjectDict(value)
        self[key] = value

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, ObjectDict):
            value = ObjectDict(value)
        super(ObjectDict, self).__setitem__(key, value)

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            if isinstance(value, dict) and not isinstance(value, ObjectDict):
                value = ObjectDict(value)
            self[key] = value

    def merge(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            if isinstance(value, dict):
                if not isinstance(value, ObjectDict):
                    value = ObjectDict(value)
                if key in self and isinstance(self[key], ObjectDict):
                    self[key].merge(value)
                    continue
            self[key] = value
